<?php
/**
 * Created by IntelliJ IDEA.
 * User: goldie
 * Date: 29/06/2016
 * Time: 18:55
 */

namespace BrewPixel\Barista\Percolator;


class FileWrite {
	/**
	 * Writes directives to .gitignore
	 *
	 * @param Composer_Ignore $gitignore
	 * @param SelfTestExceptions $exs
	 * @throws FilesystemOperationException with S/FTP form if it can't get the required filesystem credentials
	 * @throws FileOperationException
	 */
	private function rules_core_add($config, $exs) {
		$path = w3_get_pgcache_rules_core_path();
		if ($path === false)
			return;

		$original_data = @file_get_contents($path);
		if ($original_data === false)
			$original_data = '';

		$data = $original_data;

		if ($has_legacy = w3_has_rules($data, W3TC_MARKER_BEGIN_PGCACHE_LEGACY, W3TC_MARKER_END_PGCACHE_LEGACY))
			$data = w3_erase_rules($data, W3TC_MARKER_BEGIN_PGCACHE_LEGACY, W3TC_MARKER_END_PGCACHE_LEGACY);

		if ($has_wpsc = w3_has_rules($data, W3TC_MARKER_BEGIN_PGCACHE_WPSC, W3TC_MARKER_END_PGCACHE_WPSC))
			$data = w3_erase_rules($data, W3TC_MARKER_BEGIN_PGCACHE_WPSC, W3TC_MARKER_END_PGCACHE_WPSC);

		$rules = $this->rules_core_generate($config);
		$rules_missing = (strstr(w3_clean_rules($data), w3_clean_rules($rules)) === false);


		if (!$has_legacy && !$has_wpsc && !$rules_missing)
			return; // modification of file not required


		$replace_start = strpos($data, W3TC_MARKER_BEGIN_PGCACHE_CORE);
		$replace_end = strpos($data, W3TC_MARKER_END_PGCACHE_CORE);

		if ($replace_start !== false && $replace_end !== false && $replace_start < $replace_end) {
			$replace_length = $replace_end - $replace_start +
			                  strlen(W3TC_MARKER_END_PGCACHE_CORE) + 1;
		} else {
			$replace_start = false;
			$replace_length = 0;

			$search = array(
				W3TC_MARKER_BEGIN_BROWSERCACHE_NO404WP => 0,
				W3TC_MARKER_BEGIN_WORDPRESS => 0,
				W3TC_MARKER_END_MINIFY_CORE =>
					strlen(W3TC_MARKER_END_MINIFY_CORE) + 1,
				W3TC_MARKER_END_BROWSERCACHE_CACHE =>
					strlen(W3TC_MARKER_END_BROWSERCACHE_CACHE) + 1,
				W3TC_MARKER_END_PGCACHE_CACHE =>
					strlen(W3TC_MARKER_END_PGCACHE_CACHE) + 1,
				W3TC_MARKER_END_MINIFY_CACHE =>
					strlen(W3TC_MARKER_END_MINIFY_CACHE) + 1
			);

			foreach ($search as $string => $length) {
				$replace_start = strpos($data, $string);

				if ($replace_start !== false) {
					$replace_start += $length;
					break;
				}
			}
		}

		if ($replace_start !== false) {
			$data = w3_trim_rules(substr_replace($data, $rules,
				$replace_start, $replace_length));
		} else {
			$data = w3_trim_rules($data . $rules);
		}

		try {
			bp_wp_write_to_file($path, $data);
		} catch (FilesystemOperationException $ex) {
			if ($has_legacy)
				$exs->push(new FilesystemModifyException(
					$ex->getMessage(), $ex->credentials_form(),
					sprintf(__('Edit file <strong>%s</strong> and remove all lines between and including <strong>
                                %s</strong> and <strong>%s</strong> markers.', 'w3-total-cache')
						, $path
						, W3TC_MARKER_BEGIN_PGCACHE_LEGACY
						, W3TC_MARKER_END_PGCACHE_LEGACY
					), $path));
			if ($has_wpsc)
				$exs->push(new FilesystemModifyException(
					$ex->getMessage(), $ex->credentials_form(),
					sprintf(__('Edit file <strong>%s</strong> and remove all lines between and including
                                <strong>%s</strong> and <strong>%s</strong> markers.', 'w3-total-cache')
						, $path
						, W3TC_MARKER_BEGIN_PGCACHE_WPSC
						, W3TC_MARKER_END_PGCACHE_WPSC
					), $path));

			if ($rules_missing) {
				if ($replace_start !== false)
					$exs->push(new FilesystemModifyException(
						$ex->getMessage(), $ex->credentials_form(),
						sprintf(__('Edit file <strong>%s</strong> and replace all lines between and including
                                    <strong>%s</strong> and <strong>%s</strong> markers with:','w3-total-cache')
							, $path
							, W3TC_MARKER_BEGIN_PGCACHE_CORE
							, W3TC_MARKER_END_PGCACHE_CORE
						), $path, $rules));
				else
					$exs->push(new FilesystemModifyException(
						$ex->getMessage(), $ex->credentials_form(),
						sprintf(__('Edit file <strong>%s</strong> and add the following rules above the WordPress
                                    directives:')
							, $path
						), $path, $rules));
			}
		}
	}

	/**
	 * Tries to write file content
	 *
	 * @param string $filename path to file
	 * @param string $content data to write
	 * @param string $method Which method to use when creating
	 * @param string $url Where to redirect after creation
	 * @param bool|string $context folder in which to write file
	 * @throws FilesystemWriteException
	 * @return void
	 */
	function bp_wp_write_to_file($filename, $content) {
		if (@file_put_contents($filename, $content))
			return;

		try {
			w3_wp_request_filesystem_credentials();
		} catch (FilesystemOperationException $ex) {
			throw new FilesystemWriteException($ex->getMessage(),
				$ex->credentials_form(), $filename, $content);
		}

		global $wp_filesystem;
		if (!$wp_filesystem->put_contents($filename, $content)) {
			throw new FilesystemWriteException(
				'FTP credentials don\'t allow to write to file <strong>' .
				$filename . '</strong>', w3_get_filesystem_credentials_form(),
				$filename, $content);
		}
	}

	/**
	 * Returns path of pagecache core rules file
	 *
	 * @return string
	 */
	function w3_get_pgcache_rules_core_path() {
		switch (true) {
			case w3_is_apache():
			case w3_is_litespeed():
				return w3_get_home_root() . '/.htaccess';

			case w3_is_nginx():
				return w3_get_nginx_rules_path();
		}

		return false;
	}

	/**
	 * Generates rules for WP dir
	 *
	 * @param W3_Config $config
	 * @return string
	 */
	private function rules_core_generate($config) {
		switch (true) {
			case w3_is_apache():
			case w3_is_litespeed():
				return $this->rules_core_generate_apache($config);

			case w3_is_nginx():
				return $this->rules_core_generate_nginx($config);
		}

		return '';
	}

	/**
	 * Generates rules for WP dir
	 *
	 * @param W3_Config $config
	 * @return string
	 */
	private function rules_core_generate_apache($config) {
		$is_network = w3_is_network();

		$base_path = w3_get_base_path();
		$home_path = w3_get_home_path();
		$rewrite_base = ($is_network ? $base_path : $home_path);
		$cache_dir = w3_path(W3TC_CACHE_PAGE_ENHANCED_DIR);
		$permalink_structure = get_option('permalink_structure');

		$current_user = get_currentuserinfo();

		/**
		 * Auto reject cookies
		 */
		$reject_cookies = array(
			'comment_author',
			'wp-postpass'
		);

		if ($config->get_string('pgcache.engine') == 'file_generic') {
			$reject_cookies[] = 'w3tc_logged_out';
		}

		/**
		 * Reject cache for logged in users
		 * OR
		 * Reject cache for roles if any
		 */
		if ($config->get_boolean('pgcache.reject.logged')) {
			$reject_cookies = array_merge($reject_cookies, array(
				'wordpress_logged_in'
			));
		} elseif($config->get_boolean('pgcache.reject.logged_roles')) {
			$new_cookies = array();
			foreach( $config->get_array('pgcache.reject.roles') as $role ) {
				$new_cookies[] = 'w3tc_logged_' . md5(NONCE_KEY . $role);
			}
			$reject_cookies = array_merge($reject_cookies, $new_cookies);
		}

		/**
		 * Custom config
		 */
		$reject_cookies = array_merge($reject_cookies, $config->get_array('pgcache.reject.cookie'));
		w3_array_trim($reject_cookies);

		$reject_user_agents = $config->get_array('pgcache.reject.ua');
		if ($config->get_boolean('pgcache.compatibility')) {
			$reject_user_agents = array_merge(array(W3TC_POWERED_BY), $reject_user_agents);
		}

		w3_array_trim($reject_user_agents);

		/**
		 * Generate directives
		 */
		$env_W3TC_UA = '';
		$env_W3TC_REF = '';
		$env_W3TC_SSL = '';
		$env_W3TC_ENC = '';

		$rules = '';
		$rules .= W3TC_MARKER_BEGIN_PGCACHE_CORE . "\n";
		$rules .= "<IfModule mod_rewrite.c>\n";
		$rules .= "    RewriteEngine On\n";
		$rules .= "    RewriteBase " . $rewrite_base . "\n";


		if ($config->get_boolean('pgcache.debug')) {
			$rules .= "    RewriteRule ^(.*\\/)?w3tc_rewrite_test/?$ $1?w3tc_rewrite_test=1 [L]\n";
		}

		/**
		 * Check for mobile redirect
		 */
		if ($config->get_boolean('mobile.enabled')) {
			$mobile_groups = $config->get_array('mobile.rgroups');

			foreach ($mobile_groups as $mobile_group => $mobile_config) {
				$mobile_enabled = (isset($mobile_config['enabled']) ? (boolean) $mobile_config['enabled'] : false);
				$mobile_agents = (isset($mobile_config['agents']) ? (array) $mobile_config['agents'] : '');
				$mobile_redirect = (isset($mobile_config['redirect']) ? $mobile_config['redirect'] : '');

				if ($mobile_enabled && count($mobile_agents) && $mobile_redirect) {
					$rules .= "    RewriteCond %{HTTP_USER_AGENT} (" . implode('|', $mobile_agents) . ") [NC]\n";
					$rules .= "    RewriteRule .* " . $mobile_redirect . " [R,L]\n";
				}
			}
		}

		/**
		 * Check for referrer redirect
		 */
		if ($config->get_boolean('referrer.enabled')) {
			$referrer_groups = $config->get_array('referrer.rgroups');

			foreach ($referrer_groups as $referrer_group => $referrer_config) {
				$referrer_enabled = (isset($referrer_config['enabled']) ? (boolean) $referrer_config['enabled'] : false);
				$referrer_referrers = (isset($referrer_config['referrers']) ? (array) $referrer_config['referrers'] : '');
				$referrer_redirect = (isset($referrer_config['redirect']) ? $referrer_config['redirect'] : '');

				if ($referrer_enabled && count($referrer_referrers) && $referrer_redirect) {
					$rules .= "    RewriteCond %{HTTP_COOKIE} w3tc_referrer=.*(" . implode('|', $referrer_referrers) . ") [NC]\n";
					$rules .= "    RewriteRule .* " . $referrer_redirect . " [R,L]\n";
				}
			}
		}

		/**
		 * Set mobile groups
		 */
		if ($config->get_boolean('mobile.enabled')) {
			$mobile_groups = array_reverse($config->get_array('mobile.rgroups'));

			foreach ($mobile_groups as $mobile_group => $mobile_config) {
				$mobile_enabled = (isset($mobile_config['enabled']) ? (boolean) $mobile_config['enabled'] : false);
				$mobile_agents = (isset($mobile_config['agents']) ? (array) $mobile_config['agents'] : '');
				$mobile_redirect = (isset($mobile_config['redirect']) ? $mobile_config['redirect'] : '');

				if ($mobile_enabled && count($mobile_agents) && !$mobile_redirect) {
					$rules .= "    RewriteCond %{HTTP_USER_AGENT} (" . implode('|', $mobile_agents) . ") [NC]\n";
					$rules .= "    RewriteRule .* - [E=W3TC_UA:_" . $mobile_group . "]\n";
					$env_W3TC_UA = '%{ENV:W3TC_UA}';
				}
			}
		}

		/**
		 * Set referrer groups
		 */
		if ($config->get_boolean('referrer.enabled')) {
			$referrer_groups = array_reverse($config->get_array('referrer.rgroups'));

			foreach ($referrer_groups as $referrer_group => $referrer_config) {
				$referrer_enabled = (isset($referrer_config['enabled']) ? (boolean) $referrer_config['enabled'] : false);
				$referrer_referrers = (isset($referrer_config['referrers']) ? (array) $referrer_config['referrers'] : '');
				$referrer_redirect = (isset($referrer_config['redirect']) ? $referrer_config['redirect'] : '');

				if ($referrer_enabled && count($referrer_referrers) && !$referrer_redirect) {
					$rules .= "    RewriteCond %{HTTP_COOKIE} w3tc_referrer=.*(" . implode('|', $referrer_referrers) . ") [NC]\n";
					$rules .= "    RewriteRule .* - [E=W3TC_REF:_" . $referrer_group . "]\n";
					$env_W3TC_REF = '%{ENV:W3TC_REF}';
				}
			}
		}

		/**
		 * Set HTTPS
		 */
		if ($config->get_boolean('pgcache.cache.ssl')) {
			$rules .= "    RewriteCond %{HTTPS} =on\n";
			$rules .= "    RewriteRule .* - [E=W3TC_SSL:_ssl]\n";
			$rules .= "    RewriteCond %{SERVER_PORT} =443\n";
			$rules .= "    RewriteRule .* - [E=W3TC_SSL:_ssl]\n";
			$env_W3TC_SSL = '%{ENV:W3TC_SSL}';
		}

		$cache_path = str_replace(w3_get_document_root(), '', $cache_dir);

		/**
		 * Set Accept-Encoding
		 */
		if ($config->get_boolean('browsercache.enabled') && $config->get_boolean('browsercache.html.compression')) {
			$rules .= "    RewriteCond %{HTTP:Accept-Encoding} gzip\n";
			$rules .= "    RewriteRule .* - [E=W3TC_ENC:_gzip]\n";
			$env_W3TC_ENC = '%{ENV:W3TC_ENC}';
		}
		$rules .= "    RewriteCond %{HTTP_COOKIE} w3tc_preview [NC]\n";
		$rules .= "    RewriteRule .* - [E=W3TC_PREVIEW:_preview]\n";
		$env_W3TC_PREVIEW = '%{ENV:W3TC_PREVIEW}';

		$use_cache_rules = '';
		/**
		 * Don't accept POSTs
		 */
		$use_cache_rules .= "    RewriteCond %{REQUEST_METHOD} !=POST\n";

		/**
		 * Query string should be empty
		 */
		$use_cache_rules .= "    RewriteCond %{QUERY_STRING} =\"\"\n";

		/**
		 * Check permalink structure trailing slash
		 */
		if (substr($permalink_structure, -1) == '/') {
			$use_cache_rules .= "    RewriteCond %{REQUEST_URI} \\/$\n";
		}

		/**
		 * Check for rejected cookies
		 */
		$use_cache_rules .= "    RewriteCond %{HTTP_COOKIE} !(" . implode('|', array_map('w3_preg_quote', $reject_cookies)) . ") [NC]\n";

		/**
		 * Check for rejected user agents
		 */
		if (count($reject_user_agents)) {
			$use_cache_rules .= "    RewriteCond %{HTTP_USER_AGENT} !(" . implode('|', array_map('w3_preg_quote', $reject_user_agents)) . ") [NC]\n";
		}

		/**
		 * Make final rewrites for specific files
		 */
		$uri_prefix =  $cache_path . '/%{HTTP_HOST}/%{REQUEST_URI}/' .
		               '_index' . $env_W3TC_UA . $env_W3TC_REF . $env_W3TC_SSL . $env_W3TC_PREVIEW;
		$switch = " -" . ($config->get_boolean('pgcache.file.nfs') ? 'F' : 'f');

		// support for GoDaddy servers configuration which uses
		// SUBDOMAIN_DOCUMENT_ROOT variable
		if (isset($_SERVER['SUBDOMAIN_DOCUMENT_ROOT']) &&
		    $_SERVER['SUBDOMAIN_DOCUMENT_ROOT'] != $_SERVER['DOCUMENT_ROOT'])
			$document_root = '%{ENV:SUBDOMAIN_DOCUMENT_ROOT}';
		elseif (isset($_SERVER['PHP_DOCUMENT_ROOT']) &&
		        $_SERVER['PHP_DOCUMENT_ROOT'] != $_SERVER['DOCUMENT_ROOT'])
			$document_root = '%{ENV:PHP_DOCUMENT_ROOT}';
		else
			$document_root = '%{DOCUMENT_ROOT}';

		// write rule to rewrite to .html file
		$ext = '.html';
		$rules .= $use_cache_rules;
		$rules .= "    RewriteCond \"" . $document_root . $uri_prefix . $ext .
		          $env_W3TC_ENC . "\"" . $switch . "\n";
		$rules .= "    RewriteRule .* \"" . $uri_prefix . $ext .
		          $env_W3TC_ENC . "\" [L]\n";

		$rules .= "</IfModule>\n";

		$rules .= W3TC_MARKER_END_PGCACHE_CORE . "\n";

		return $rules;
	}
}
